#pragma once

#include "easy.h"

#include "slist.inl"
#include "global.inl"
#include "exception.inl"

namespace cppcurl
{

    namespace __impl
    {

        /* opt_writer */

        template<typename T_value, typename = void>
        struct opt_writer;

        template<typename T>
        struct opt_writer<
            T,
            std::enable_if_t<
                std::is_same_v<bool, std::decay_t<T>>>>
        {
            static inline void write(CURL* handle, CURLoption opt, bool val)
            {
                easy_exception::throw_error(
                    curl_easy_setopt(handle, opt, cppcurl_convert_cast<long>(val)),
                    "Unable to set option on easy handle.");
            }
        };

        template<typename T>
        struct opt_writer<
            T,
            std::enable_if_t<
                std::is_same_v<long, std::decay_t<T>>>>
        {
            static inline void write(CURL* handle, CURLoption opt, long val)
            {
                easy_exception::throw_error(
                    curl_easy_setopt(handle, opt, val),
                    "Unable to set option on easy handle.");
            }
        };

        template<typename T>
        struct opt_writer<
            T,
            std::enable_if_t<
                std::is_same_v<void*, std::decay_t<T>>>>
        {
            static inline void write(CURL* handle, CURLoption opt, void* val)
            {
                easy_exception::throw_error(
                    curl_easy_setopt(handle, opt, val),
                    "Unable to set option on easy handle.");
            }
        };

        template<typename T>
        struct opt_writer<
            T,
            std::enable_if_t<
                std::is_same_v<data_callback, std::decay_t<T>>>>
        {
            static inline void write(CURL* handle, CURLoption opt, data_callback val)
            {
                easy_exception::throw_error(
                    curl_easy_setopt(handle, opt, val),
                    "Unable to set option on easy handle.");
            }
        };

        template<typename T>
        struct opt_writer<
            T,
            std::enable_if_t<
                std::is_same_v<std::string, std::decay_t<T>>>>
        {
            static inline void write(CURL* handle, CURLoption opt, const std::string& val)
            {
                easy_exception::throw_error(
                    curl_easy_setopt(handle, opt, val.c_str()),
                    "Unable to set option on easy handle.");
            }
        };

        template<typename T>
        struct opt_writer<
            T,
            std::enable_if_t<
                std::is_same_v<slist, std::decay_t<T>>>>
        {
            static inline void write(CURL* handle, CURLoption opt, const slist& val)
            {
                easy_exception::throw_error(
                    curl_easy_setopt(handle, opt, val.handle()),
                    "Unable to set option on easy handle.");
            }
        };

        /* info_reader */

        template<typename T_value, typename = void>
        struct info_reader;

        template<typename T>
        struct info_reader<
            T,
            std::enable_if_t<
                std::is_same_v<long, std::decay_t<T>>>>
        {
            static inline long read(CURL* handle, CURLINFO info)
            {
                long ret;
                easy_exception::throw_error(
                    curl_easy_getinfo(handle, info, &ret),
                    "Unable to get info from easy handle.");
                return ret;
            }
        };

        template<typename T>
        struct info_reader<
            T,
            std::enable_if_t<
                std::is_same_v<std::string, std::decay_t<T>>>>
        {
            static inline std::string read(CURL* handle, CURLINFO info)
            {
                const char * ret;
                easy_exception::throw_error(
                    curl_easy_getinfo(handle, info, &ret),
                    "Unable to get info from easy handle.");
                return std::string(ret ? ret : "");
            }
        };

    }

    /* easy */

    easy::easy()
        : _handle(nullptr, &curl_easy_cleanup)
    {
        if (!global::is_initialized())
            throw multi_exception(CURLM_BAD_HANDLE, "libcurl was not initialized!");
        _handle.reset(curl_easy_init());
        if (!_handle)
            throw multi_exception(CURLM_OUT_OF_MEMORY, "Unable to create easy handle");
    }

    CURL* easy::handle() const
        { return _handle.get(); }

    void easy::reset()
    {
        _header.reset();
        curl_easy_reset(_handle.get());
    }

    void easy::perform()
    {
        easy_exception::throw_error(
            curl_easy_perform(_handle.get()),
            "Unable to perform easy handle operations.");
    }

    void easy::add_header(const std::string& s)
    {
        _header.append(s);
        set_option<CURLOPT_HTTPHEADER>(_header);
    }

    template<CURLoption T_option, typename T_value>
    void easy::set_option(const __impl::option<T_option, T_value>& opt)
        {  __impl::opt_writer<T_value>::write(handle(), T_option, opt.value); }

    template<CURLoption T_option>
    void easy::set_option(const typename option<T_option>::value_type& val)
        { set_option(option<T_option>(val)); }

    template<CURLINFO T_info>
    typename info<T_info>::value_type easy::get_info() const
    {
        using info_type = info<T_info>;
        using value_type = typename info_type::value_type;
        return __impl::info_reader<value_type>::read(handle(), T_info);
    }

}
