#pragma once

#include <memory>
#include <curl/curl.h>

namespace cppcurl
{

    struct easy;

    struct multi
    {
    private:
        using handle_ptr_u = std::unique_ptr<CURLM, decltype(&curl_multi_cleanup)>;

    private:
        handle_ptr_u _handle;

    public:
        /**
         * @brief Constructor.
         */
        inline multi();

        /**
         * @brief Get the libcurl multi handle.
         */
        inline CURLM * handle() const;

        /**
         * @brief Add a standard curl handle to the multi stack.
         */
        inline void add_handle(const easy& e);

        /**
         * @brief Removes a curl handle from the multi stack again.
         */
        inline void remove_handle(const easy& e);

        /**
         * @brief Write all file descriptors known by the object to the passed file descriptor sets.
         *
         * @param[in] read      File descriptor set to listen for read operations.
         * @param[in] write     File descriptor set to listen for write operations.
         * @param[in] except    File descriptor set to listen for exceptional condition.
         *
         * @return              Greatest file descriptor.
         */
        inline int prepare_fdsets(fd_set& read, fd_set& write, fd_set& except) const;

        /**
         * @brief Perform any operations on the multi handle.
         *
         * @return Number of running easy handles.
         */
        inline int perform();

        /**
         * @brief Get the max timeout to wait for file descriptors.
         *
         * @return Timeout to wait for file descriptors or -1 if not relevant.
         */
        inline long timeout() const;

        /**
         * @brief Read pending messages from the queue.
         */
        inline CURLMsg * read_info() const;

        /**
         * @brief Read all pending messages from the queue.
         *        Call the passed predicate one for each message.
         */
        template<typename T_pred>
        inline void read_infos(const T_pred&& pred) const;
    };

}

#include "multi.inl"
